import json
from unittest2 import TestCase

from werkzeug.wrappers import Response
from werkzeug.test import Client as TestClient

from cosmic.api import API
from cosmic.http import Server
from cosmic.models import BaseModel
from cosmic.globals import cosmos
from cosmic.types import *


cookbook_spec = {
    u'name': u'cookbook',
    u'actions': {
        u'map': {
            u'cabbage': {
                u'accepts': {
                    u'type': u"Struct",
                    u"param": {
                        u"map": {
                            u"spicy": {
                                u"required": True,
                                u"schema": {u"type": u"Boolean"}
                            },
                            u"capitalize": {
                                u"required": False,
                                u"schema": {u"type": u"Boolean"}
                            }
                        },
                        u"order": [u"spicy", u"capitalize"]
                    }
                },
                u'returns': {u'type': u'String'},
                u'doc': u"Yay cabbage"
            },
            u'noop': {
                u'doc': u"Does not do anything"
            }
        },
        u'order': [u'cabbage', u'noop']
    },
    u"models": {
        u"map": {
            u"Recipe": {
                u"properties": {
                    u"map": {
                        u"name": {
                            u"required": True,
                            u"schema": {u"type": u"String"}
                        },
                    },
                    u"order": [u"name"]
                },
                u"links": {
                    u"map": {},
                    u"order": []
                },
                u"query_fields": {
                    u"map": {},
                    u"order": []
                },
                u"list_metadata": {
                    u"map": {},
                    u"order": []
                },
                u'methods': {
                    u'get_by_id': False,
                    u'get_list': False,
                    u'create': False,
                    u'update': False,
                    u'delete': False,
                },
            },
            u"Author": {
                u"properties": {
                    u"map": {
                        u"is_gordon_ramsay": {
                            u"required": True,
                            u"schema": {u"type": u"Boolean"}
                        },
                    },
                    u"order": [u"is_gordon_ramsay"]
                },
                u"links": {
                    u"map": {},
                    u"order": []
                },
                u"query_fields": {
                    u"map": {
                        u"is_gordon_ramsay": {
                            u"required": True,
                            u"schema": {u"type": u"Boolean"}
                        },
                        },
                    u"order": [u"is_gordon_ramsay"]
                },
                u"list_metadata": {
                    u"map": {},
                    u"order": []
                },
                u'methods': {
                    u'get_by_id': False,
                    u'get_list': True,
                    u'create': False,
                    u'update': False,
                    u'delete': False,
                },
            }
        },
        u"order": [u"Recipe", u"Author"]
    }
}


class TestAPI(TestCase):

    def setUp(self):
        self.maxDiff = None

        self._old_cosmos = cosmos.data
        cosmos.data = {}

        self.cookbook = cookbook = API(u'cookbook')

        @cookbook.action(
            accepts=Struct([
                required(u"spicy", Boolean),
                optional(u"capitalize", Boolean)
            ]),
            returns=String)
        def cabbage(spicy, capitalize=False):
            "Yay cabbage"
            if spicy:
                c = "kimchi"
            else:
                c = "sauerkraut"
            if capitalize:
                return c.capitalize()
            else:
                return c

        @cookbook.action(accepts=None, returns=None)
        def noop():
            "Does not do anything"
            pass

        @cookbook.model
        class Recipe(BaseModel):
            properties = [
                required(u"name", String)
            ]

            @classmethod
            def validate_patch(cls, datum):
                if datum["name"] == "bacon":
                    raise ValidationError("Not kosher")

        @cookbook.model
        class Author(BaseModel):
            methods = ['get_list']
            properties = [
                required(u"is_gordon_ramsay", Boolean)
            ]
            query_fields = [
                required(u"is_gordon_ramsay", Boolean)
            ]
            @classmethod
            def get_list(cls, is_gordon_ramsey):
                return [("0", {"is_gordon_ramsey": True})]

        self.Author = Author

        self.server = Server(self.cookbook)
        self.server.debug = True
        self.app = self.server.wsgi_app
        self.client = TestClient(self.app, response_wrapper=Response)

    def tearDown(self):
        cosmos.data = self._old_cosmos

    def test_get_list_missing(self):
        resp = self.client.get('/Author')
        self.assertEqual(resp.status_code, 400)

    def test_model(self):
        d = {
            "_links": {
                "self": {"href": "/Recipe/24"}
            },
            "name": "pancake"
        }

        (id, rep) = Representation(Model('cookbook.Recipe')).from_json(d)
        self.assertEqual(rep['name'], "pancake")
        self.assertEqual(Representation(Model('cookbook.Recipe')).to_json((id, rep)), d)

    def test_model_deserialize_okay(self):
        (id, rep) = Representation(Model('cookbook.Recipe')).from_json({
            "_links": {
                "self": {"href": "/Recipe/14"}
            },
            "name": "turkey"
        })
        self.assertEqual(rep['name'], "turkey")

    def test_subclassing_hook(self):
        self.assertEqual(set(self.cookbook.models.__dict__.keys()), set(["Recipe", "Author"]))

    def test_recursive_subclassing_hook(self):
        @self.cookbook.model
        class ChocolateAuthor(self.Author):
            pass

        self.assertEqual(set(self.cookbook.models.__dict__.keys()), set(["Recipe", "Author", "ChocolateAuthor"]))

    def test_model_schema_validation(self):
        with self.assertRaises(ValidationError):
            Representation(Model('cookbook.Recipe')).from_json(1.1)

    def test_model_custom_validation(self):
        with self.assertRaisesRegexp(ValidationError, "kosher"):
            (id, rep) = Representation(Model('cookbook.Recipe')).from_json({
                "_links": {
                    "self": {"href": "/Recipe/123"}
                },
                "name": "bacon"
            })
            self.cookbook.models.Recipe.validate_patch(rep)

    def test_serialize(self):
        self.assertEqual(APISpec.to_json(self.cookbook.spec), cookbook_spec)

    def test_call_action_with_args(self):
        self.assertEqual(self.cookbook.actions.cabbage(spicy=False), "sauerkraut")

    def test_spec_endpoint(self):
        res = self.client.get('/spec.json')
        self.assertEqual(json.loads(res.data), cookbook_spec)

    def test_spec_wrong_method(self):
        res = self.client.get('/actions/noop')
        self.assertEqual(res.status_code, 404)
        res = self.client.post('/spec.json')
        self.assertEqual(res.status_code, 404)

    def test_wrong_content_type(self):
        res = self.client.post('/actions/cabbage', data="1", content_type="application/xml")
        self.assertEqual(res.status_code, 400)
        self.assertRegexpMatches(res.data, "Content-Type")

    def test_action_okay(self):
        data = json.dumps({"spicy": True})
        res = self.client.post('/actions/cabbage', data=data, content_type="application/json")
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.data, '"kimchi"')

    def test_noop_action_okay(self):
        res = self.client.post('/actions/noop', data='')
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, '')

    def test_schema(self):
        APISpec.from_json(APISpec.to_json(self.cookbook.spec))




from unittest2 import TestCase

from cosmic.models import BaseModel
from cosmic.globals import cosmos
from cosmic.http import *
from cosmic.api import API
from cosmic.client import WsgiAPIClient
from cosmic.types import *


class TestGuideWhatIsAPI(TestCase):
    maxDiff = None

    def setUp(self):
        self.cosmos = {}
        with cosmos.swap(self.cosmos):
            self.mathy = API("trivia", homepage="http://example.com")

    def test_to_json(self):
        with cosmos.swap(self.cosmos):
            self.assertEqual(APISpec.to_json(self.mathy.spec), {
                u'name': 'trivia',
                u'homepage': 'http://example.com',
                u'actions': {u'map': {}, u'order': []},
                u'models': {u'map': {}, u'order': []}
            })


class TestGuideModels(TestCase):
    maxDiff = None

    def setUp(self):
        self.cosmos = {}
        with cosmos.swap(self.cosmos):
            self.places = places = API('places')

            @places.model
            class Address(BaseModel):
                properties = [
                    required(u"number", Integer),
                    required(u"street", String),
                    optional(u"city", String)
                ]

    def test_to_json(self):
        places = self.places
        with cosmos.swap(self.cosmos):
            self.assertEqual(APISpec.to_json(places.spec), {
                u'name': u'places',
                u'actions': {u'map': {}, u'order': []},
                u"models": {
                    u"map": {
                        u"Address": {
                            u"properties": {
                                u"map": {
                                    u"number": {
                                        u"required": True,
                                        u"schema": {u"type": u"Integer"}
                                    },
                                    u"street": {
                                        u"required": True,
                                        u"schema": {u"type": u"String"}
                                    },
                                    u"city": {
                                        u"required": False,
                                        u"schema": {u"type": u"String"}
                                    }
                                },
                                u"order": [u"number", u"street", u"city"]
                            },
                            u"links": {u"map": {}, u"order": []},
                            u"query_fields": {u"map": {}, u"order": []},
                            u"list_metadata": {u"map": {}, u"order": []},
                            u'methods': {
                                u'get_by_id': False,
                                u'get_list': False,
                                u'create': False,
                                u'update': False,
                                u'delete': False,
                            },
                        }
                    },
                    u"order": [u"Address"]
                }
            })

    def test_serialize_model(self):
        places = self.places
        with cosmos.swap(self.cosmos):
            rep = {
                "number": 31,
                "street": "Sesame"
            }

            self.assertEqual(Representation(Model('places.Address')).to_json((None, rep)), rep)


class TestGuideModelLinks(TestCase):
    maxDiff = None

    def setUp(self):
        self.cosmos1 = {}
        with cosmos.swap(self.cosmos1):
            self.places = places = API('places')

            @places.model
            class City(BaseModel):
                properties = [
                    optional(u"name", String)
                ]

            @places.model
            class Address(BaseModel):
                properties = [
                    required(u"number", Integer),
                    required(u"street", String),
                ]
                links = [
                    required_link(u"city", Model('places.City'))
                ]

        self.cosmos2 = {}
        with cosmos.swap(self.cosmos2):

            class PlacesClient(WsgiAPIClient):
                wsgi_app = Server(places).wsgi_app

            self.remote_places = PlacesClient()

    def remote_create_models(self):
        with cosmos.swap(self.cosmos2):
            elm13 = self.remote_places.models.Address(number=13, street="Elm")
            self.assertEqual(elm13.number, 13)


class TestGuideGetById(TestCase):
    maxDiff = None

    def setUp(self):
        self.cosmos = {}
        with cosmos.swap(self.cosmos):

            self.places = places = API('places')

            @places.model
            class City(BaseModel):
                properties = [
                    optional(u"name", String)
                ]

                @classmethod
                def get_by_id(cls, id):
                    if id in cities:
                        return cities[id]
                    else:
                        return None

            cities = {
                "0": {"name": "Toronto"},
                "1": {"name": "San Francisco"},
            }

    def test_access_links(self):
        places = self.places
        with cosmos.swap(self.cosmos):
            city = places.models.City.get_by_id("0")
            self.assertEqual(city['name'], "Toronto")
            self.assertEqual(places.models.City.get_by_id("5") is None, True)


class TestGuideSave(TestCase):
    maxDiff = None

    def setUp(self):
        self.cosmos = {}
        with cosmos.swap(self.cosmos):
            self.places = places = API('places')

            @places.model
            class City(BaseModel):
                methods = ["create", "update"]
                properties = [
                    optional(u"name", String)
                ]

                @classmethod
                def validate_patch(cls, datum):
                    if datum[u"name"][0].islower():
                        raise ValidationError("Name must be capitalized", datum["name"])

                @classmethod
                def create(cls, **patch):
                    id = str(len(cities))
                    cities[id] = patch
                    return id, patch

                @classmethod
                def update(cls, id, **patch):
                    cities[id] = patch
                    return patch

        self.cosmos2 = {}
        with cosmos.swap(self.cosmos2):

            class PlacesClient(WsgiAPIClient):
                wsgi_app = Server(places).wsgi_app
                server_cosmos = self.cosmos

            self.remote_places = PlacesClient()

        cities = {
            "0": {"name": "Toronto"},
            "1": {"name": "San Francisco"},
        }

    def test_save_good(self):
        places = self.places
        with cosmos.swap(self.cosmos):
            (id, rep) = places.models.City.create(name="Moscow")
            self.assertEqual(id, "2")

    def test_local_save_validation_error(self):
        with cosmos.swap(self.cosmos):
            with self.assertRaisesRegexp(ValidationError, "must be capitalized"):
                self.places.models.City.validate_patch({"name": "moscow"})
                self.places.models.City.create(name="moscow")

    def test_remote_save_validation_error(self):
        with cosmos.swap(self.cosmos2):
            with self.assertRaisesRegexp(RemoteHTTPError, "must be capitalized"):
                self.remote_places.models.City.create(name="moscow")


class TestGuideDelete(TestCase):
    maxDiff = None

    def setUp(self):
        self.cosmos = {}
        with cosmos.swap(self.cosmos):
            self.places = places = API('places')

            @places.model
            class City(BaseModel):
                methods = ["get_by_id", "delete"]
                properties = [
                    optional(u"name", String)
                ]

                @classmethod
                def get_by_id(cls, id):
                    if id in cities:
                        return cities[id]
                    raise NotFound

                @classmethod
                def delete(cls, id):
                    del cities[id]

            cities = {
                "0": {"name": "Toronto"},
                "1": {"name": "San Francisco"},
            }

    def test_save(self):
        places = self.places
        with cosmos.swap(self.cosmos):
            city = places.models.City.get_by_id("0")
            places.models.City.delete("0")
            with self.assertRaises(NotFound):
                places.models.City.get_by_id("0")


class TestGuideGetList(TestCase):
    maxDiff = None

    def setUp(self):

        self.cosmos = {}
        with cosmos.swap(self.cosmos):

            self.places = places = API('places')

            @places.model
            class City(BaseModel):
                properties = [
                    optional(u"name", String)
                ]
                query_fiels = [
                    optional(u"country", String)
                ]

                @classmethod
                def get_list(cls, country=None):
                    if country is None:
                        return cities.items()
                    elif country == "Canada":
                        return [("0", cities["0"])]
                    elif country == "USA":
                        return [("1", cities["1"])]
                    else:
                        return []

            cities = {
                "0": {"name": "Toronto"},
                "1": {"name": "San Francisco"},
            }

    def test_get_list(self):
        places = self.places
        with cosmos.swap(self.cosmos):
            l1 = places.models.City.get_list()
            self.assertEqual(len(l1), 2)
            l2 = places.models.City.get_list(country="Canada")
            self.assertEqual(len(l2), 1)
            self.assertEqual(l2[0][1]['name'], "Toronto")
            l3 = places.models.City.get_list(country="Russia")
            self.assertEqual(l3, [])


class TestGuideAction(TestCase):
    maxDiff = None

    def setUp(self):
        self.cosmos1 = {}
        with cosmos.swap(self.cosmos1):
            self.mathy = mathy = API("mathy")

            @mathy.action(accepts=Array(Integer), returns=Integer)
            def add(numbers):
                return sum(numbers)

            @mathy.action(accepts=Struct([
                required(u'numerator', Integer),
                required(u'denominator', Integer),
            ]), returns=Integer)
            def divide(numerator, denominator):
                return numerator / denominator


            self.add = add

        self.cosmos2 = {}
        with cosmos.swap(self.cosmos2):

            class MathyClient(WsgiAPIClient):
                wsgi_app = Server(mathy).wsgi_app

            self.remote_mathy = MathyClient()

    def test_call_as_function(self):
        self.assertEqual(self.add([1, 2, 3]), 6)

    def test_call_as_action(self):
        self.assertEqual(self.mathy.actions.add([1, 2, 3]), 6)

    def test_call_as_action_remote(self):
        self.assertEqual(self.remote_mathy.actions.add([1, 2, 3]), 6)

    def test_call_as_action_remote_kwargs(self):
        self.assertEqual(self.remote_mathy.actions.divide(numerator=10, denominator=5), 2)

    def test_remote_action_validation_error(self):
        with self.assertRaisesRegexp(HTTPError, "Invalid Integer"):
            self.remote_mathy.actions.add([1, 2, True])



import copy
import json

from unittest2 import TestCase
from werkzeug.wrappers import Response
from werkzeug.test import Client as TestClient

from cosmic.client import WsgiAPIClient, ClientLoggingMixin
from cosmic.globals import cosmos
from cosmic.exceptions import *
from planetarium import *


json_spec = {
    u"name": u"planetarium",
    u"models": {
        u"map": {
            u"Sphere": {
                u"properties": {
                    u"map": {
                        u"name": {
                            u"required": True,
                            u"schema": {u"type": u"String"}
                        },
                        u"temperature": {
                            u"required": True,
                            u"schema": {u"type": u"Float"}
                        },
                    },
                    u"order": [u"name", u"temperature"],
                },
                u"list_metadata": {
                    u"map": {
                        u"last_updated": {
                            u"required": True,
                            u"schema": {u"type": u"DateTime"}
                        },
                    },
                    u"order": [u"last_updated"],

                },
                u"query_fields": {
                    u"map": {
                        u"name": {
                            u"required": False,
                            u"schema": {u"type": u"String"}
                        },
                        u"revolves_around": {
                            u"required": False,
                            u"schema": {u"type": u"String"}
                        }
                    },
                    u"order": [u"name", u"revolves_around"]
                },
                u"links": {
                    u"map": {
                        u"revolves_around": {
                            u"model": u"planetarium.Sphere",
                            u"required": False
                        }
                    },
                    u"order": [u"revolves_around"]
                },
                u'methods': {
                    u'get_by_id': True,
                    u'get_list': True,
                    u'create': True,
                    u'update': True,
                    u'delete': True,
                },
            },
        },
        u"order": [u"Sphere"]
    },
    u"actions": {
        u"map": {
            u"hello": {
                u"accepts": {u"type": u"cosmic.Representation", u"param": u"planetarium.Sphere"},
                u"returns": {u"type": u"String"}
            }
        },
        u"order": [u"hello"]
    }
}


class Retry(Exception):
    pass


class PlanetariumClient(ClientLoggingMixin, WsgiAPIClient):
    token = None
    wsgi_app = planetarium_server.wsgi_app

    def build_request(self, endpoint, *args, **kwargs):
        request = super(PlanetariumClient, self).build_request(endpoint, *args, **kwargs)
        if self.token is not None:
            request.headers["X-Danish"] = self.token
        return request

    def parse_response(self, endpoint, res):
        if res.status_code == 401:
            raise Retry()
        return super(PlanetariumClient, self).parse_response(endpoint, res)

    def call(self, endpoint, *args, **kwargs):
        while True:
            try:
                return super(PlanetariumClient, self).call(endpoint, *args, **kwargs)
            except Retry:
                # Find token, e.g. through OAuth
                self.token = "poppyseed"
                continue

class _LocalRemotePlanetariumTest(TestCase):
    maxDiff = None

    def setUp(self):

        self.local_cosmos = dict(cosmos.items())
        self.remote_cosmos = dict()

        with cosmos.swap(self.remote_cosmos):
            PlanetariumClient()

    def _test(self):
        raise NotImplementedError()

    def _test_logs(self):
        pass

    def test_local(self):
        with cosmos.swap(self.local_cosmos):
            self._test()

    def test_remote(self):
        with cosmos.swap(self.remote_cosmos):
            self._test()
            self._test_logs()


class TestGetById(_LocalRemotePlanetariumTest):

    def _test(self):
        with db.swap(planet_db):
            Sphere = cosmos['planetarium'].models.Sphere
            with self.assertRaises(NotFound):
                Sphere.get_by_id('100')
            rep = Sphere.get_by_id('0')
            self.assertEqual(rep['name'], "Sun")

    def _test_logs(self):

        (req, res) = cosmos['planetarium'].log.pop()

        self.assertEqual(req["method"], "GET")
        self.assertEqual(req["url"], "/Sphere/0")
        self.assertEqual(req["data"], "")

        self.assertEqual(res["status_code"], 200)
        self.assertEqual(dict(res["headers"])["Content-Type"], "application/json")
        self.assertEqual(json.loads(res["data"]), {
            u"name": u"Sun",
            u"temperature": 3000.0,
            "_links": {
                "self": {"href": "/Sphere/0"}
            },
        })

        (req, res) = cosmos['planetarium'].log.pop()

        self.assertEqual(req["method"], "GET")
        self.assertEqual(req["url"], "/Sphere/100")
        self.assertEqual(req["data"], "")
        self.assertEqual(res["status_code"], 404)


class TestCreate(_LocalRemotePlanetariumTest):

    def _test(self):
        with db.swap(copy.deepcopy(planet_db)):
            Sphere = cosmos['planetarium'].models.Sphere

            (id, rep) = Sphere.create(
                name="Pluto",
                revolves_around="0")

            self.assertEqual(id, "3")

            full = {
                "name": "Pluto",
                "temperature": 60,
                "_links": {
                    "self": {"href": "/Sphere/3"},
                    "revolves_around": {"href": "/Sphere/0"},
                }
            }
            self.assertEqual(Representation(Model('planetarium.Sphere')).to_json((id, rep)), full)
            self.assertEqual(db['Sphere']['3'], {
                'name': 'Pluto',
                'temperature': 60,
                'revolves_around': '0',
            })


class TestUpdateLink(_LocalRemotePlanetariumTest):

    def _test(self):
        with db.swap(copy.deepcopy(planet_db)):
            Sphere = cosmos['planetarium'].models.Sphere
            # Save property
            moon = Sphere.get_by_id("2")
            self.assertEqual(moon['revolves_around'], "1")

            rep = Sphere.update("2",
                                name=moon['name'],
                                revolves_around="0")

            self.assertEqual(rep['revolves_around'], "0")
            self.assertEqual(db['Sphere']['2'], {
                'name': moon['name'],
                'temperature': -50,
                'revolves_around': '0',
            })

    def _test_logs(self):

        (req, res) = cosmos['planetarium'].log.pop()

        patch = {
            u'_links': {
                u'self': {u'href': u'/Sphere/2'},
                u'revolves_around': {u'href': u'/Sphere/0'},
            },
            u'name': u'Moon',
        }

        updated = {
            u'_links': {
                u'self': {u'href': u'/Sphere/2'},
                u'revolves_around': {u'href': u'/Sphere/0'},
            },
            u'temperature': -50,
            u'name': u'Moon',
        }

        self.assertEqual(req["method"], "PUT")
        self.assertEqual(req["url"], "/Sphere/2")
        self.assertEqual(json.loads(req["data"]), patch)
        self.assertEqual(dict(req["headers"])["Content-Type"], "application/json")

        self.assertEqual(res["status_code"], 200)
        self.assertEqual(json.loads(res["data"]), updated)
        self.assertEqual(dict(res["headers"])["Content-Type"], "application/json")


class TestUpdateProperty(_LocalRemotePlanetariumTest):

    def _test(self):
        with db.swap(copy.deepcopy(planet_db)):
            Sphere = cosmos['planetarium'].models.Sphere
            moon = Sphere.get_by_id("2")

            rep = Sphere.update("2",
                                name="Luna",
                                revolves_around=moon['revolves_around'])

            self.assertEqual(rep['name'], "Luna")
            self.assertEqual(db['Sphere']['2'], {
                'name': 'Luna',
                'temperature': -50,
                'revolves_around': moon['revolves_around'],
            })

    def _test_logs(self):

        (req, res) = cosmos['planetarium'].log.pop()

        patch = {
            u'_links': {
                u'self': {u'href': u'/Sphere/2'},
                u'revolves_around': {u'href': u'/Sphere/1'},
            },
            u'name': u'Luna',
        }

        updated = {
            u'_links': {
                u'self': {u'href': u'/Sphere/2'},
                u'revolves_around': {u'href': u'/Sphere/1'},
            },
            u'temperature': -50,
            u'name': u'Luna',
        }

        self.assertEqual(req["method"], "PUT")
        self.assertEqual(req["url"], "/Sphere/2")
        self.assertEqual(json.loads(req["data"]), patch)
        self.assertEqual(dict(req["headers"])["Content-Type"], "application/json")

        self.assertEqual(res["status_code"], 200)
        self.assertEqual(json.loads(res["data"]), updated)
        self.assertEqual(dict(res["headers"])["Content-Type"], "application/json")


class TestGetList(_LocalRemotePlanetariumTest):

    def _test(self):
        with db.swap(planet_db):
            Sphere = cosmos['planetarium'].models.Sphere
            res = Sphere.get_list(name="Oops")
            self.assertEqual(len(res[0]), 0)
            res = Sphere.get_list(name="Sun")
            self.assertEqual(len(res[0]), 1)
            self.assertEqual(res[0][0][0], "0")
            res = Sphere.get_list()
            self.assertEqual(len(res[0]), 3)

    def _test_logs(self):

        (req, res) = cosmos['planetarium'].log.pop()

        self.assertEqual(req["method"], "GET")
        self.assertEqual(req["url"], "/Sphere")
        self.assertEqual(req["data"], "")

        self.assertEqual(res["status_code"], 200)
        self.assertEqual(json.loads(res["data"]), {
            "_links": {
                "self": {"href": "/Sphere"}
            },
            "_embedded": {
                "Sphere": [
                    {
                        '_links': {'self': {'href': '/Sphere/0'}},
                        'name': 'Sun',
                        'temperature': 3000.0,
                        },
                    {
                        '_links': {
                            'revolves_around': {'href': '/Sphere/0'},
                            'self': {'href': '/Sphere/1'}
                        },
                        'name': 'Earth',
                        'temperature': 30.0,
                        },
                    {
                        '_links': {
                            'revolves_around': {'href': '/Sphere/1'},
                            'self': {'href': '/Sphere/2'}
                        },
                        'name': 'Moon',
                        'temperature': -50.0,
                        },
                    ]
            },
            'last_updated': '2014-01-01T00:00:00'
        })
        self.assertEqual(dict(res["headers"])["Content-Type"], "application/json")

        (req, res) = cosmos['planetarium'].log.pop()

        url = "/Sphere?name=Sun"
        self.assertEqual(req["method"], "GET")
        self.assertEqual(req["url"], url)
        self.assertEqual(req["data"], "")

        self.assertEqual(res["status_code"], 200)
        self.assertEqual(json.loads(res["data"]), {
            "_links": {
                "self": {"href": "/Sphere?name=Sun"}
            },
            "_embedded": {
                "Sphere": [
                    {
                        '_links': {'self': {'href': '/Sphere/0'}},
                        'name': 'Sun',
                        'temperature': 3000.0,
                        },
                    ]
            },
            'last_updated': '2014-01-01T00:00:00'
        })
        self.assertEqual(dict(res["headers"])["Content-Type"], "application/json")


class TestDelete(_LocalRemotePlanetariumTest):

    def _test(self):
        with db.swap(copy.deepcopy(planet_db)):
            Sphere = cosmos['planetarium'].models.Sphere
            Sphere.delete("1")
            self.assertTrue('1' not in db['Sphere'])
            with self.assertRaises(NotFound):
                Sphere.delete("123811")

    def _test_logs(self):

        (req, res) = cosmos['planetarium'].log.pop()
        self.assertEqual(res["status_code"], 404)

        (req, res) = cosmos['planetarium'].log.pop()

        self.assertEqual(req["method"], "DELETE")
        self.assertEqual(req["url"], "/Sphere/1")
        self.assertEqual(req["data"], "")

        self.assertEqual(res["status_code"], 204)
        self.assertEqual(res["data"], "")


class TestAction(_LocalRemotePlanetariumTest):

    def _test(self):
        pluto = {
            "name": "Pluto",
            "temperature": -120,
        }
        self.assertEqual(cosmos['planetarium'].actions.hello((None, pluto)), "Hello, Pluto")


class TestPlanitarium(TestCase):
    maxDiff = None

    def test_spec_endpoint(self):
        app = planetarium_server.wsgi_app
        client = TestClient(app, response_wrapper=Response)
        res = client.get('/spec.json', headers={'X-Danish': 'poppyseed'})
        self.assertEqual(json.loads(res.data), json_spec)




from multiprocessing import Process
from contextlib import contextmanager

from .models import BaseModel
from .exceptions import NotFound
from .globals import SwappableDict, cosmos


db = SwappableDict()


class DBModel(BaseModel):
    @classmethod
    def get_by_id(cls, id):
        try:
            return db[cls.table_name][id]
        except KeyError:
            raise NotFound

    @classmethod
    def get_list(cls, **kwargs):
        if not kwargs:
            return sorted(db[cls.table_name].items())
        ret = []
        for id, rep in db[cls.table_name].items():
            keep = True
            for key, val in kwargs.items():
                if rep[key] != val:
                    keep = False
                    break
            if keep:
                ret.append((id, rep))
        return sorted(ret)

    @classmethod
    def create(cls, **patch):
        id = str(len(db[cls.table_name]))
        db[cls.table_name][id] = patch
        return id, patch

    @classmethod
    def update(cls, id, **patch):
        rep = db[cls.table_name][id]
        rep.update(patch)
        return rep

    @classmethod
    def delete(cls, id):
        if id not in db[cls.table_name]:
            raise NotFound
        del db[cls.table_name][id]


@contextmanager
def served_api(api, port, **kwargs):
    from time import sleep

    p = Process(target=api.run, args=(port,), kwargs=kwargs)
    p.start()
    sleep(0.1)

    try:
        with cosmos.swap({}):
            yield
    except Exception as e:
        raise e
    finally:
        p.terminate()


class WildcardClass(object):
    def __eq__(self, other):
        return True

Wildcard = WildcardClass()




from time import sleep

from cosmic.api import API
from cosmic.http import Server
from cosmic.types import *
from cosmic.models import BaseModel
from cosmic.exceptions import NotFound

words = API('words')


@words.action(accepts=String, returns=String)
def pluralize(word):
    if word.endswith('y'):
        return word[:-1] + 'ies'
    else:
        return word + 's'


@words.model
class Word(BaseModel):
    methods = ['get_by_id']
    properties = [
        required(u"letters", String)
    ]
    @classmethod
    def get_by_id(cls, id):
        if id == "0":
            return {"letters": "hello"}
        else:
            raise NotFound

@words.action(accepts=Integer)
def lock_thread(seconds):
    sleep(seconds)

@words.action(accepts=Representation(Model('words.Word')), returns=Integer)
def count_letters(word):
    return len(word[1]['letters'])

wsgi_app = Server(words).wsgi_app

if __name__ == "__main__":
    words.run(use_reloader=True)



from time import sleep
from unittest2 import TestCase
from threading import Thread

from cosmic.globals import *
from cosmic.globals import storage

class TestSwappableDict(TestCase):

    def test_dict_methods(self):
        s = SwappableDict()
        s['a'] = 1
        self.assertEqual(s['a'], 1)
        self.assertEqual(len(s), 1)
        del s['a']
        self.assertEqual(len(s), 0)

    def test_swap(self):
        s = SwappableDict()
        s.update({'a': 1})
        self.assertEqual(dict(s), {'a': 1})
        with s.swap({'b': 2}):
            self.assertEqual(dict(s), {'b': 2})
        self.assertEqual(dict(s), {'a': 1})

class TestThreadLocalDict(TestCase):

    def test_unbound(self):
        s = ThreadLocalDict()
        with self.assertRaises(Exception):
            s['a'] = 1

    def test_threads(self):

        s = ThreadLocalDict()

        def run():
            sleep(0.2)
            with thread_local():
                self.assertEqual(dict(s), {})
                s['a'] = 1
                self.assertEqual(dict(s), {'a': 1})
                del s['a']
                self.assertEqual(dict(s), {})
                sleep(0.2)

        threads = []
        for i in range(10):
            thread = Thread(target=run)
            thread.start()
            threads.append(thread)

        for thread in threads:
            thread.join()

        self.assertEqual(storage, {})



from cosmic.api import API
from cosmic.http import Server
from cosmic.types import *
from cosmic.models import BaseModel
from cosmic.exceptions import NotFound

quotes = API('quotes')

data = {
    "0": {"text": "Know thyself.", "author": "Socrates"},
    "1": {"text": "Never give up.", "author": "Shuzo Matsuoka"}
}

@quotes.model
class Quote(BaseModel):
    methods = ['get_by_id']
    properties = [
        required("text", String)
    ]
    @classmethod
    def get_by_id(cls, id):
        if id in data:
            return data[id]
        else:
            raise NotFound

wsgi_app = Server(words).wsgi_app

if __name__ == "__main__":
    words.run(use_reloader=True)



from datetime import datetime

from unittest2 import TestCase

from cosmic.types import *


class TestURLParams(TestCase):
    def setUp(self):
        self.schema = URLParams([
            optional("foo", String),
            required("bars", Array(Integer))
        ])

    def test_okay(self):
        self.assertEqual(self.schema.from_json('foo=Wha&bars=[1]'), {
            "foo": "Wha",
            "bars": [1]
        })
        self.assertEqual(self.schema.from_json('foo=Wha&bars=[1,2,3]'), {
            "foo": "Wha",
            "bars": [1, 2, 3]
        })
        self.assertEqual(self.schema.from_json('bars=[1,2,3]'), {
            "bars": [1, 2, 3]
        })

    def test_string_wrapper(self):
        schema = URLParams([
            optional("birthday", DateTime)
        ])
        d = datetime(year=1991, month=8, day=12)
        self.assertEqual(schema.to_json({"birthday": d}), "birthday=1991-08-12T00%3A00%3A00")
        self.assertEqual(schema.from_json("birthday=1991-08-12T00%3A00%3A00"), {"birthday": d})

    def test_wrong_deep_type(self):
        with self.assertRaisesRegexp(ValidationError, "Invalid Integer"):
            self.schema.from_json('foo=Wha&bars=[1,1.2]')

    def test_unexpected_repeat(self):
        with self.assertRaises(ValidationError):
            self.schema.from_json('foo=Wha&bars=[1]&foo=Bing')





import json
from collections import OrderedDict

from werkzeug.urls import url_decode, url_encode
from werkzeug.datastructures import MultiDict
from .legacy_teleport import standard_types, ParametrizedWrapper, BasicWrapper, \
    required, optional, Box, ValidationError

from .globals import cosmos

__all__ = ['Integer', 'Float', 'Boolean', 'String', 'Binary', 'DateTime',
           'JSON', 'Array', 'Map', 'OrderedMap', 'Struct', 'Schema', 'Model',
           'Link', 'Representation', 'Patch', 'APISpec', 'URLParams',
           'Box', 'required', 'optional', 'required_link', 'optional_link',
           'ValidationError']


def getter(name):
    if name == "cosmic.APISpec":
        return APISpec
    elif name == "cosmic.Model":
        return Model
    elif name == "cosmic.Link":
        return Link
    elif name == "cosmic.Representation":
        return Representation
    elif name == "cosmic.Patch":
        return Patch
    else:
        raise KeyError()

# Explicit definitions for static analyzers
Integer = None
Float = None
Boolean = None
String = None
Binary = None
DateTime = None
JSON = None
Array = None
Map = None
OrderedMap = None
Struct = None
Schema = None
globals().update(standard_types(getter))


def required_link(name, model, doc=None):
    return name, {"model": model, "required": True, "doc": doc}


def optional_link(name, model, doc=None):
    return name, {"model": model, "required": False, "doc": doc}


class Model(BasicWrapper):
    """A Teleport type representing an API model. Its JSON form is a dotted
    string, the native form is an instance of this class.
    """
    type_name = "cosmic.Model"
    schema = String

    def __init__(self, full_name):
        self.full_name = full_name
        self.api_name, self.model_name = full_name.split('.', 1)

    @property
    def model_spec(self):
        try:
            return cosmos[self.api_name].spec['models'][self.model_name]
        except KeyError:
            raise RuntimeError('Model does not exist: {}'.format(self.full_name))

    @classmethod
    def assemble(cls, datum):
        if '.' in datum:
            a, m = datum.split('.', 1)
            if len(a) and len(m):
                return cls(datum)
        raise ValidationError('Invalid Model: {}'.format(datum))

    @classmethod
    def disassemble(cls, datum):
        return '{}.{}'.format(datum.api_name, datum.model_name)


def link_in_href(model):
    return Struct([
        required(u"href", Link(model))
    ])



class Link(ParametrizedWrapper):
    """A Teleport type representing a link to an object. Its native form is
    simply a resource id. It takes a :class:`~cosmic.types.Model` as
    parameter:

    .. code:: python

        >>> Link(Model('places.City')).to_json("3")
        "/City/3"
        >>> Link(Model('places.City')).from_json("/City/3")
        "3"

    """
    type_name = "cosmic.Link"
    param_schema = Model

    def __init__(self, param):
        self.param = param
        self.schema = String

    def assemble(self, datum):
        parts = datum.split('/')
        if parts[-2] != self.param.model_name:
            raise ValidationError("Invalid url for %s link: %s" % (self.param.model_name, datum))
        return parts[-1]

    def disassemble(self, datum):
        return "/%s/%s" % (self.param.model_name, datum)


class BaseRepresentation(ParametrizedWrapper):
    param_schema = Model

    def __init__(self, param):
        self.param = param
        self._lazy_schema = None

    @property
    def schema(self):
        if self._lazy_schema is None:

            links = [
                ("self", {
                    "required": False,
                    "schema": Struct([
                        required(u"href", Link(self.param))
                    ])
                })
            ]
            for name, link in self.param.model_spec['links'].items():
                is_required = False
                if not self.all_fields_optional:
                    is_required = link["required"]

                links.append((name, {
                    "required": is_required,
                    "schema": Struct([
                        required(u"href", Link(link['model']))
                    ])
                }))

            props = [
                optional("_links", Struct(links)),
            ]
            for name, field in self.param.model_spec['properties'].items():
                is_required = False
                if not self.all_fields_optional:
                    is_required = field["required"]

                props.append((name, {
                    "required": is_required,
                    "schema": field['schema'],
                }))

            self._lazy_schema = Struct(props)

        return self._lazy_schema

    def assemble(self, datum):
        rep = {}
        links = {}
        for name, hlink in datum.pop("_links", {}).items():
            links[name] = hlink['href']

        self_id = links.pop("self", None)

        rep.update(links)
        rep.update(datum)

        return self_id, rep


    def disassemble(self, datum):
        (id, rep) = datum

        links = {}
        if id is not None:
            links["self"] = {'href': id}
        for name, link in self.param.model_spec['links'].items():
            value = rep.get(name, None)
            if value != None:
                links[name] = {
                    "href": value
                }
        d = {}
        if links:
            d["_links"] = links
        for name in self.param.model_spec['properties'].keys():
            value = rep.get(name, None)
            if value != None:
                d[name] = value

        return d


class Representation(BaseRepresentation):
    """A Teleport type representing a model representation. Its native form is
    a dict as defined by :data:`~cosmic.models.BaseModel.properties` and
    :data:`~cosmic.models.BaseModel.links`. Links are represented by plain
    string ids.

    It takes a :class:`~cosmic.types.Model` as parameter.
    """
    type_name = "cosmic.Representation"
    all_fields_optional = False


class Patch(BaseRepresentation):
    """A Teleport type representing a model patch. Its native form is similar
    to that of :class:`~cosmic.types.Representation`, except all fields are
    optional. To make a field required, use
    :meth:`~cosmic.models.BaseModel.validate_patch`.

    It takes a :class:`~cosmic.types.Model` as parameter.
    """
    type_name = "cosmic.Patch"
    all_fields_optional = True

    def assemble(self, datum):
        self_id, rep = super(Patch, self).assemble(datum)
        model_obj = getattr(cosmos[self.param.api_name].models, self.param.model_name)
        model_obj.validate_patch(rep)
        return self_id, rep


class URLParams(ParametrizedWrapper):
    """A Teleport type that behaves mostly like the :class:`Struct`
    type, except it serializes the data into a query string:

    .. code:: python

        >>> p = URLParams([
        ...     required("foo", Boolean),
        ...     required("bar", Integer)
        ... ])
        ...
        >>> p.to_json({"foo": True, "bar": 3})
        'foo=true&bar=3'
        >>> p.from_json("foo=false&bar=0")
        {'foo': False, 'bar': 0}

    A string parameter or a parameter whose type is a wrapper over string will
    not require quotes:

    .. code:: python

        >>> from cosmic.types import DateTime
        >>> schema = URLParams([
        ...     required('birthday', DateTime)
        ... ])
        >>> schema.from_json('birthday=1991-08-12T00%3A00%3A00')
        {'birthday': datetime.datetime(1991, 8, 12, 0, 0)}

    """
    schema = String

    def __init__(self, param):
        if type(param) == list:
            param = OrderedDict(param)
        self.param = param

    def assemble(self, datum):
        # Use Werkzeug to turn URL params into a dict
        return self.from_multi_dict(url_decode(datum))

    def disassemble(self, datum):
        return url_encode(self.to_multi_dict(datum))

    def from_multi_dict(self, md):
        from .tools import is_string_type
        # Where only a single param was
        md = md.to_dict(flat=False)
        ret = {}
        for name, field in self.param.items():
            if name in md.keys():
                if len(md[name]) > 1:
                    raise ValidationError("Repeating query parameters not allowed: %s" % name)
                if is_string_type(field['schema']):
                    ret[name] = md[name][0]
                else:
                    ret[name] = json.loads(md[name][0])
        return Struct(self.param).from_json(ret)

    def to_multi_dict(self, datum):
        from .tools import is_string_type

        d = Struct(self.param).to_json(datum)
        md = MultiDict()
        for name, field in self.param.items():
            if name in d.keys():
                if is_string_type(field['schema']):
                    md[name] = d[name]
                else:
                    md[name] = json.dumps(d[name])
        return md


class APISpec(BasicWrapper):
    """The teleport type that encapsulates all metadata associated with an API.
    This type is used to prepare the output for the ``/spec.json`` endpoint,
    as well as to deserialize it when building an API client.

    What's inside? Well, if you insist:

    .. code:: python

        Struct([
            required("name", String),
            optional("homepage", String),
            required("actions", OrderedMap(Struct([
                optional("accepts", Schema),
                optional("returns", Schema),
                optional("doc", String)
            ]))),
            required("models", OrderedMap(Struct([
                required("properties", OrderedMap(Struct([
                    required("schema", Schema),
                    required("required", Boolean),
                    optional("doc", String)
                ]))),
                required("links", OrderedMap(Struct([
                    required("model", Model),
                    required("required", Boolean),
                    optional("doc", String)
                ]))),
                required("query_fields", OrderedMap(Struct([
                    required("schema", Schema),
                    required("required", Boolean),
                    optional("doc", String)
                ]))),
                required("methods", Struct([
                    required("get_by_id", Boolean),
                    required("get_list", Boolean),
                    required("create", Boolean),
                    required("update", Boolean),
                    required("delete", Boolean),
                    ])),
                required("list_metadata", OrderedMap(Struct([
                    required("schema", Schema),
                    required("required", Boolean),
                    optional("doc", String)
                ])))
            ])))
        ])

    """
    type_name = "cosmic.APISpec"

    schema = Struct([
        required("name", String),
        optional("homepage", String),
        required("actions", OrderedMap(Struct([
            optional("accepts", Schema),
            optional("returns", Schema),
            optional("doc", String)
        ]))),
        required("models", OrderedMap(Struct([
            required("properties", OrderedMap(Struct([
                required("schema", Schema),
                required("required", Boolean),
                optional("doc", String)
            ]))),
            required("links", OrderedMap(Struct([
                required("model", Model),
                required("required", Boolean),
                optional("doc", String)
            ]))),
            required("query_fields", OrderedMap(Struct([
                required("schema", Schema),
                required("required", Boolean),
                optional("doc", String)
            ]))),
            required("methods", Struct([
                required("get_by_id", Boolean),
                required("get_list", Boolean),
                required("create", Boolean),
                required("update", Boolean),
                required("delete", Boolean),
                ])),
            required("list_metadata", OrderedMap(Struct([
                required("schema", Schema),
                required("required", Boolean),
                optional("doc", String)
            ])))
        ])))
    ])

    @classmethod
    def assemble(cls, datum):
        from .tools import validate_underscore_identifier

        for model_name, model_spec in datum['models'].items():
            link_names = set(model_spec['links'].keys())
            field_names = set(model_spec['properties'].keys())

            if link_names & field_names:
                raise ValidationError(
                    "Model cannot contain a field and link with the same name: {}".format(model_name))

            for name in link_names | field_names:
                validate_underscore_identifier(name)

            if 'id' in link_names | field_names:
                raise ValidationError("'id' is a reserved name.")

        return datum





from contextlib import contextmanager
from collections import MutableMapping
from werkzeug.local import get_ident

from cosmic.exceptions import ThreadLocalMissing

__all__ = ['thread_local', 'SwappableDict', 'ThreadLocalDict']

storage = {}


@contextmanager
def thread_local():
    """A context manager for safely creating and deleting the thread-local
    necessary for :class:`~cosmic.globals.ThreadLocalDict`.

    .. code::

        g = ThreadLocalDict()

        with thread_local():
            # g is only accessible within this context
            g['foo'] = 1

    """
    ident = get_ident()
    storage[ident] = local = {}
    yield local
    del storage[ident]


@contextmanager
def ensure_thread_local():
    if get_ident() in storage:
        yield
    else:
        with thread_local():
            yield


def thread_local_middleware(app):
    """To put your entire application in a :func:`~cosmic.globals.thread_local`
    context, you must put it at the entry point of your application's thread.
    In the case of a threading WSGI server, this is achieved by wrapping your
    WSGI application in this middleware.

    Note that just like the rest of this module, this middleware isn't tied to
    Cosmic in any way. For example, you can use it to enable a
    :class:`~cosmic.globals.ThreadLocalDict` inside a Django application:

    .. code::

        from django.core.wsgi import get_wsgi_application
        from cosmic.globals import thread_local_middleware

        application = thread_local_middleware(get_wsgi_application())

    .. seealso::
        `How to deploy Django with WSGI
        <https://docs.djangoproject.com/en/dev/howto/deployment/wsgi/>`_

    :param app: WSGI application
    :return: WSGI application
    """
    def application(environ, start_response):
        with thread_local():
            return app(environ, start_response)
    return application


class SwappableDict(MutableMapping):

    def __init__(self, data=None):
        if data is None:
            self.data = {}
        else:
            self.data = data

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, val):
        self.data[key] = val

    def __delitem__(self, key):
        del self.data[key]

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return self.data.__iter__()

    def keys(self):
        return self.data.keys()

    def __repr__(self):
        return repr(self.data)

    @contextmanager
    def swap(self, new):
        old = self.data
        self.data = new
        yield
        self.data = old


class ThreadLocalDict(SwappableDict):
    """A dictionary-like object that stores its values inside a thread-local.
    Useful for storing global state safely. Instantiate this once and import
    it everywhere to read and write global values (e.g. the currently
    authenticated user or database connections).
    """

    def __init__(self):
        pass

    def __repr__(self):
        if get_ident() in storage:
            return repr(self.data)
        else:
            return '<unbound ThreadLocalDict>'

    @property
    def data(self):
        try:
            local = storage[get_ident()]
        except KeyError:
            raise ThreadLocalMissing()
        return local.setdefault(id(self), {})


cosmos = SwappableDict()



from __future__ import unicode_literals

import re
import inspect
import json

from .exceptions import SpecError
from .types import *


__all__ = ['get_args', 'args_to_datum', 'assert_is_compatible',
           'deserialize_json', 'serialize_json', 'string_to_json',
           'validate_underscore_identifier', 'is_string_type']


def get_args(func):
    """Given a function, returns a tuple (*required*, *optional*), tuples of
    non-keyword and keyword arguments respectively. If a function contains
    splats (\* or \**), a :exc:`~cosmic.exceptions.SpecError` will be raised.
    """
    args, varargs, keywords, defaults = inspect.getargspec(func)
    if varargs or keywords:
        raise SpecError("Cannot define action with splats (* or **)")
    # No arguments
    if len(args) == 0:
        return (), ()
    # Number of non-keyword arguments (required ones)
    numargs = len(args) - (len(defaults) if defaults else 0)
    required_args = tuple(args[:numargs])
    optional_args = tuple(args[numargs:])
    return required_args, optional_args


def args_to_datum(*args, **kwargs):
    """Takes arbitrary args and kwargs and packs them into a dict if there are
    more than one. Returns `None` if there are no arguments. Must be called
    with either a single argument or multiple keyword arguments.
    """
    if len(args) == 1 and len(kwargs) == 0:
        return args[0]
    if len(args) == 0 and len(kwargs) > 0:
        return kwargs
    if len(args) == 0 and len(kwargs) == 0:
        return None
    raise SpecError("Action must be called either with one argument or with one or more keyword arguments")


def assert_is_compatible(schema, required_args, optional_args):
    """Raises a :exc:`~cosmic.exceptions.SpecError` if function argument spec
    (as returned by :func:`get_args`) is incompatible with the given schema.
    By incompatible, it is meant that there exists such a piece of data that
    is valid according to the schema, but that could not be applied to the
    function by :func:`apply_to_func`.
    """
    # No arguments
    if len(required_args + optional_args) == 0:
        raise SpecError("Function needs to accept arguments")
    # One argument can accept anything
    if len(required_args + optional_args) == 1:
        return
    # Multiple arguments means schema must be Struct
    if not isinstance(schema, Struct):
        raise SpecError("For a function that takes arguments, accepts schema"
                        " is expected to be a Struct")
    # Each non-keyword argument in the function must have a corresponding
    # required field in the schema
    for r in required_args:
        if r not in schema.param.keys() or not schema.param[r]["required"]:
            raise SpecError("Action argument '%s' must have a corresponding"
                            " required field in the accepts schema" % r)
    # All fields in the schema must have a corresponding function argument
    for f in schema.param.keys():
        if f not in set(required_args + optional_args):
            raise SpecError("The '%s' field must have a corresponding"
                            " function argument" % f)


def deserialize_json(schema, datum):
    if schema and not datum:
        raise ValidationError("Expected Box, found None")
    if datum and not schema:
        raise ValidationError("Expected None, found Box")
    if schema and datum:
        return schema.from_json(datum.datum)
    return None


def serialize_json(schema, datum):
    if schema is not None and datum is None:
        raise ValidationError("Expected data, found None")
    if datum is not None and schema is None:
        raise ValidationError("Expected None, found data")
    if schema is not None and datum is not None:
        return Box(schema.to_json(datum))
    return None


def string_to_json(s):
    if s == "":
        return None
    else:
        return Box(json.loads(s))


def validate_underscore_identifier(id):
    if re.match('^[a-zA-Z0-9_]*$', id) is None:
        raise SpecError("Identifier must consist of [a-zA-Z0-9_] characters only: %s" % id)
    if id.startswith('_') or id.endswith('_'):
        raise SpecError("Identifier cannot start or end with an underscore: %s" % id)
    if '__' in id:
        raise SpecError("Identifier cannot have consecutive underscores: %s" % id)


def is_string_type(serializer):
    if Schema.to_json(serializer)['type'] == 'String':
        return True
    if hasattr(serializer, 'schema'):
        return is_string_type(serializer.schema)
    return False



import json

import requests
from werkzeug.exceptions import NotFound as WerkzeugNotFound
from werkzeug.wrappers import Request, Response
from werkzeug.routing import Rule
from werkzeug.routing import Map as RuleMap

from .types import *
from .tools import get_args, string_to_json, args_to_datum, deserialize_json, \
    serialize_json
from .exceptions import *
from .globals import ensure_thread_local


class Server(object):
    url_map = RuleMap([
        Rule('/spec.json', endpoint='spec', methods=['GET']),
        Rule('/actions/<action>', endpoint='action', methods=['POST']),
        Rule('/<model>/<id>', endpoint='get_by_id', methods=['GET']),
        Rule('/<model>/<id>', endpoint='update', methods=['PUT']),
        Rule('/<model>/<id>', endpoint='delete', methods=['DELETE']),
        Rule('/<model>', endpoint='create', methods=['POST']),
        Rule('/<model>', endpoint='get_list', methods=['GET']),
    ])

    def __init__(self, api, debug=False):
        self.api = api
        self.debug = debug

    def dispatch_request(self, request):
        adapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint_name, values = adapter.match()
        except WerkzeugNotFound:
            return error_response("Not Found", 404)

        if endpoint_name == 'spec':
            endpoint = SpecEndpoint(self.api.spec)

        elif endpoint_name == 'action':
            action_name = values.pop('action')
            if action_name not in self.api.spec['actions'].keys():
                return error_response("Not Found", 404)
            endpoint = ActionEndpoint(
                self.api.spec,
                action_name,
                getattr(self.api.actions, action_name))
        else:
            model_name = values.pop('model')
            if model_name not in self.api.spec['models'].keys():
                return error_response("Not Found", 404)
            model_spec = self.api.spec['models'][model_name]
            if not model_spec['methods'][endpoint_name]:
                return error_response("Method Not Allowed", 405)
            model_obj = getattr(self.api.models, model_name)
            endpoints = {
                'get_by_id': GetByIdEndpoint,
                'create': CreateEndpoint,
                'update': UpdateEndpoint,
                'delete': DeleteEndpoint,
                'get_list': GetListEndpoint,
            }
            args = {
                "api_spec": self.api.spec,
                "model_name": model_name,
                "func": getattr(model_obj, endpoint_name),
            }
            endpoint = endpoints[endpoint_name](**args)

        try:
            return self.view(endpoint, request, **values)
        except HTTPError as err:
            return error_response(err.message, err.code)
        except ValidationError as err:
            return error_response(str(err), 400)

    def wsgi_app(self, environ, start_response):
        with ensure_thread_local():
            request = Request(environ)
            if self.debug:
                response = self.dispatch_request(request)
            else:
                try:
                    response = self.dispatch_request(request)
                except Exception as exc:
                    response = self.unhandled_exception_hook(exc, request)

            return response(environ, start_response)

    def unhandled_exception_hook(self, exc, request):
        return error_response("Internal Server Error", 500)

    def view(self, endpoint, request, **url_args):
        func_input = self.parse_request(endpoint, request, **url_args)
        func_output = endpoint.handler(**func_input)
        return self.build_response(endpoint,
                                   func_input=func_input,
                                   func_output=func_output)

    def parse_request(self, endpoint, request, **url_args):
        return endpoint.parse_request(request, **url_args)

    def build_response(self, endpoint, func_input, func_output):
        return endpoint.build_response(func_input=func_input,
                                       func_output=func_output)





def error_response(message, code):
    body = json.dumps({"error": message})
    return Response(body, code, {"Content-Type": "application/json"})


def get_payload_from_http_message(req):
    bytes = req.data
    if not bytes:
        return None
    if req.mimetype != "application/json":
        raise SpecError('Content-Type must be "application/json" got "%s" instead' % req.mimetype)
    charset = req.mimetype_params.get("charset", "utf-8")
    if charset.lower() != "utf-8":
        raise SpecError('Content-Type charset must be "utf-8" got %s instead' % charset)
    try:
        data = bytes.decode('utf-8')
    except UnicodeDecodeError:
        raise SpecError("Unicode Decode Error")
    try:
        return Box(json.loads(data))
    except ValueError:
        raise SpecError("Invalid JSON")


def reverse_werkzeug_url(url, values):
    rule = Rule(url)
    # Rule needs to be bound before building
    RuleMap([rule])
    # Note: this seems to be changing in Werkzeug master
    domain_part, url = rule.build(values)
    return url


class Endpoint(object):
    response_can_be_empty = True
    response_must_be_empty = None

    request_can_be_empty = True
    request_must_be_empty = None

    query_schema = None

    acceptable_exceptions = []

    def handler(self, *args, **kwargs):
        if not self.acceptable_exceptions:
            return self.func(*args, **kwargs)

        try:
            return Either(value=self.func(*args, **kwargs))
        except tuple(self.acceptable_exceptions) as e:
            return Either(exception=e)

    def parse_request(self, request, **url_args):
        req = {
            'url_args': url_args,
            'headers': request.headers
        }
        try:
            req['json'] = get_payload_from_http_message(request)
        except SpecError as e:
            raise HTTPError(code=400, message=e.args[0])

        is_empty = request.data == ""

        if ((self.request_must_be_empty == True and not is_empty) or
                (is_empty and self.request_can_be_empty == False)):
            raise HTTPError(code=400, message="Invalid data")

        if self.query_schema is not None:
            req['query'] = self.query_schema.from_multi_dict(request.args)

        return req

    def build_response(self, func_input, func_output):
        raise NotImplementedError()

    def build_request(self, data=None, url_args=None, headers=None, query=None):

        if url_args is None:
            url_args = {}
        if headers is None:
            headers = {}
        if query is None:
            query = {}

        if data is not None:
            headers["Content-Type"] = "application/json"
            string_data = json.dumps(data.datum)
        else:
            string_data = ""

        url = reverse_werkzeug_url(self.url, url_args)

        if self.query_schema is not None and query:
            query_string = self.query_schema.to_json(query)
            if query_string:
                url += "?%s" % query_string

        return requests.Request(
            method=self.method,
            url=url,
            data=string_data,
            headers=headers)

    def parse_response(self, res):

        is_empty = res.text == ""
        if ((self.response_must_be_empty == True and not is_empty) or
                (is_empty and self.response_can_be_empty == False)):
            raise SpecError("Invalid response")

        r = {
            'code': res.status_code,
            'headers': res.headers
        }

        try:
            r['json'] = string_to_json(res.text)
        except ValueError:
            raise SpecError("Unparseable response")

        if r['code'] not in self.acceptable_response_codes:
            message = None
            if 'json' in r and r['json'] and type(r['json'].datum) == dict and 'error' in r['json'].datum:
                message = r['json'].datum['error']

            raise RemoteHTTPError(code=r['code'], message=message)

        return r


class SpecEndpoint(Endpoint):
    """
    :Request:
        :Method: ``GET``
        :URL: ``/spec.json``

    :Response:
        :Code: ``200``
        :Body: The API spec as a JSON-encoded string.
        :ContentType: ``application/json``

    """
    method = "GET"
    acceptable_response_codes = [200]

    def __init__(self, api_spec=None):
        self.url = '/spec.json'
        self.api_spec = api_spec

    def parse_request(self, req, **url_args):
        return {}

    def build_request(self, *args, **kwargs):
        return super(SpecEndpoint, self).build_request()

    def handler(self):
        return self.api_spec

    def parse_response(self, res):
        res = super(SpecEndpoint, self).parse_response(res)
        return APISpec.from_json(res['json'].datum)

    def build_response(self, func_input, func_output):
        body = json.dumps(APISpec.to_json(func_output))
        return Response(body, 200, {"Content-Type": "application/json"})


class ActionEndpoint(Endpoint):
    """
    :Request:
        :Method: ``POST``
        :URL: ``/actions/<action>`` where *action* is the action name.
        :Body: The action parameters as a JSON-encoded string or empty if the
            action expects no parameters.
        :ContentType: ``application/json`` if body is not empty.
    :Response:
        :Code: ``200`` or ``204`` if body is empty.
        :Body: The return value as a JSON-encoded string or empty if the actions
            has no return value.
        :ContentType: ``application/json`` if body is not empty.

    """

    method = "POST"
    acceptable_response_codes = [200, 204]

    def __init__(self, api_spec, action_name, func=None):
        self.func = func
        self.action_name = action_name
        self.action_spec = api_spec['actions'][action_name]
        self.accepts = self.action_spec.get('accepts', None)
        self.returns = self.action_spec.get('returns', None)
        self.url = "/actions/%s" % action_name

    def build_request(self, *args, **kwargs):
        packed = args_to_datum(*args, **kwargs)
        data = serialize_json(self.accepts, packed)
        return super(ActionEndpoint, self).build_request(data=data)

    def parse_request(self, req, **url_args):
        req = super(ActionEndpoint, self).parse_request(req, **url_args)
        data = deserialize_json(self.accepts, req['json'])
        kwargs = {}
        if data is not None:
            required_args, optional_args = get_args(self.func)
            # If only one argument, take the whole object
            if len(required_args + optional_args) == 1:
                kwargs = {required_args[0]: data}
            else:
                kwargs = data
        return kwargs

    def parse_response(self, res):
        res = super(ActionEndpoint, self).parse_response(res)
        if self.returns and res['json']:
            return self.returns.from_json(res['json'].datum)
        else:
            return None

    def build_response(self, func_input, func_output):
        data = serialize_json(self.returns, func_output)
        if data is None:
            return Response("", 204, {})
        else:
            body = json.dumps(data.datum)
            return Response(body, 200, {"Content-Type": "application/json"})


class GetByIdEndpoint(Endpoint):
    """
    :Request:
        :Method: ``GET``
        :URL: ``/<model>/<id>`` where *model* is the model name.
    :Response:
        :Code: ``200`` or ``404`` if object is not found.
        :Body: The object representation as a JSON-encoded string.
        :ContentType: ``application/json`` if body is not empty.

    """
    method = "GET"
    acceptable_response_codes = [404, 200]
    response_can_be_empty = True
    request_must_be_empty = True
    acceptable_exceptions = [NotFound]

    def __init__(self, api_spec, model_name, func=None):
        self.model_name = model_name
        self.full_model_name = "{}.{}".format(api_spec['name'], model_name)
        self.func = func
        self.url = "/%s/<id>" % model_name

    def build_request(self, id):
        return super(GetByIdEndpoint, self).build_request(
            url_args={'id': id})

    def parse_request(self, req, **url_args):
        req = super(GetByIdEndpoint, self).parse_request(req, **url_args)
        return {'id': req['url_args']['id']}

    def build_response(self, func_input, func_output):
        if func_output.exception is not None:
            return Response("", 404, {})
        else:
            id = func_input['id']
            rep = func_output.value
            body = json.dumps(Representation(Model(self.full_model_name)).to_json((id, rep)))
            return Response(body, 200, {"Content-Type": "application/json"})

    def parse_response(self, res):
        res = super(GetByIdEndpoint, self).parse_response(res)
        if res['code'] == 404:
            raise NotFound
        if res['code'] == 200:
            (id, rep) = Representation(Model(self.full_model_name)).from_json(res['json'].datum)
            return rep


class UpdateEndpoint(Endpoint):
    """
    :Request:
        :Method: ``PUT``
        :URL: ``/<model>/<id>`` where *model* is the model name.
        :Body: A corresponding model patch as a JSON-encoded string.
        :ContentType: ``application/json``
    :Response:
        :Code: ``200`` or ``404`` if object is not found.
        :Body: New model representation as a JSON-encoded string.
        :ContentType: ``application/json``

    """
    method = "PUT"
    acceptable_response_codes = [200, 404]
    response_can_be_empty = False
    request_can_be_empty = False
    acceptable_exceptions = [NotFound]

    def __init__(self, api_spec, model_name, func=None):
        self.model_name = model_name
        self.full_model_name = "{}.{}".format(api_spec['name'], model_name)
        self.func = func
        self.url = "/%s/<id>" % model_name

    def build_request(self, id, **patch):
        return super(UpdateEndpoint, self).build_request(
            data=Box(Patch(Model(self.full_model_name)).to_json((id, patch))),
            url_args={'id': id})

    def parse_request(self, req, **url_args):
        req = super(UpdateEndpoint, self).parse_request(req, **url_args)
        id, rep = Patch(Model(self.full_model_name)).from_json(req['json'].datum)
        rep['id'] = req['url_args']['id']
        return rep

    def build_response(self, func_input, func_output):
        if func_output.exception is not None:
            return Response("", 404, {})
        else:
            id = func_input['id']
            rep = func_output.value
            body = json.dumps(Representation(Model(self.full_model_name)).to_json((id, rep)))
            return Response(body, 200, {"Content-Type": "application/json"})

    def parse_response(self, res):
        res = super(UpdateEndpoint, self).parse_response(res)
        if res['code'] == 200:
            return Representation(Model(self.full_model_name)).from_json(res['json'].datum)[1]
        if res['code'] == 404:
            raise NotFound


class CreateEndpoint(Endpoint):
    """
    :Request:
        :Method: ``POST``
        :URL: ``/<model>`` where *model* is the model name.
        :Body: Corresponding model patch as a JSON-encoded string.
        :ContentType: ``application/json``
    :Response:
        :Code: ``201``
        :Body: New model representation as a JSON-encoded string.
        :ContentType: ``application/json``

    """
    method = "POST"
    acceptable_response_codes = [201]
    response_can_be_empty = False
    request_can_be_empty = False

    def __init__(self, api_spec, model_name, func=None):
        self.model_name = model_name
        self.full_model_name = "{}.{}".format(api_spec['name'], model_name)
        self.func = func
        self.url = "/%s" % model_name

    def build_request(self, **patch):
        return super(CreateEndpoint, self).build_request(
            data=Box(Patch(Model(self.full_model_name)).to_json((None, patch))))

    def parse_request(self, req, **url_args):
        req = super(CreateEndpoint, self).parse_request(req, **url_args)
        id, rep = Patch(Model(self.full_model_name)).from_json(req['json'].datum)
        return rep

    def parse_response(self, res):
        res = super(CreateEndpoint, self).parse_response(res)
        return Representation(Model(self.full_model_name)).from_json(res['json'].datum)

    def build_response(self, func_input, func_output):
        body = json.dumps(Representation(Model(self.full_model_name)).to_json(func_output))
        href = "/%s/%s" % (self.model_name, func_output[0])
        return Response(body, 201, {
            "Location": href,
            "Content-Type": "application/json"
        })


class DeleteEndpoint(Endpoint):
    """
    :Request:
        :Method: ``DELETE``
        :URL: ``/<model>/<id>`` where *model* is the model name.
    :Response:
        :Code: ``204`` or ``404`` if object is not found.
        :Body: Empty.

    """
    method = "DELETE"
    acceptable_response_codes = [204, 404]
    response_must_be_empty = True
    request_must_be_empty = True
    acceptable_exceptions = [NotFound]

    def __init__(self, api_spec, model_name, func=None):
        self.model_name = model_name
        self.full_model_name = "{}.{}".format(api_spec['name'], model_name)
        self.func = func
        self.url = "/%s/<id>" % model_name

    def build_request(self, id):
        return super(DeleteEndpoint, self).build_request(
            url_args={'id': id})

    def parse_request(self, req, **url_args):
        req = super(DeleteEndpoint, self).parse_request(req, **url_args)
        return {'id': req['url_args']['id']}

    def parse_response(self, res):
        res = super(DeleteEndpoint, self).parse_response(res)
        if res['code'] == 204:
            return None
        if res['code'] == 404:
            raise NotFound

    def build_response(self, func_input, func_output):
        if func_output.exception is not None:
            return Response("", 404, {})
        else:
            return Response("", 204, {})


class GetListEndpoint(Endpoint):
    """
    :Request:
        :Method: ``GET``
        :URL: ``/<model>`` where *model* is the model name.
        :Query: Query parameters serialized by the model's *query_schema*
    :Response:
        :Code: ``200``
        :ContentType: ``application/json``
        :Body:

            The syntax follows `JSON HAL
            <http://stateless.co/hal_specification.html>`_ specification.

            .. code::

                {
                    "_links": {
                        "self": {"href": <self>}
                    },
                    "_embedded": {
                        <model>: [<repr>*]
                    },
                    <metadata>*
                }

            In the above, *self* is the request URL, *model* is the name of
            the model that was requested and *repr* is a JSON-encoded
            representation of an object matched by the query. *Metadata* is
            defined according to
            :data:`~cosmic.models.BaseModel.list_metadata`.

    """
    method = "GET"
    acceptable_response_codes = [200]
    response_can_be_empty = False
    request_must_be_empty = True

    def __init__(self, api_spec, model_name, func=None):
        self.model_name = model_name
        self.full_model_name = "{}.{}".format(api_spec['name'], model_name)
        self.model_spec = api_spec['models'][model_name]
        self.list_metadata = self.model_spec['list_metadata']
        self.func = func
        self.query_schema = None
        if self.model_spec['query_fields']:
            self.query_schema = URLParams(self.model_spec['query_fields'])
        self.url = "/%s" % model_name

    def build_request(self, **query):
        return super(GetListEndpoint, self).build_request(query=query)

    def parse_request(self, req, **url_args):
        req = super(GetListEndpoint, self).parse_request(req, **url_args)
        return req.get('query', {})

    def parse_response(self, res):
        res = super(GetListEndpoint, self).parse_response(res)
        j = res['json'].datum
        l = []
        for jrep in j["_embedded"][self.model_name]:
            l.append(Representation(Model(self.full_model_name)).from_json(jrep))

        if self.list_metadata:
            meta = j.copy()
            del meta['_embedded']
            del meta['_links']
            meta = Struct(self.list_metadata).from_json(meta)
            return l, meta
        return l

    def build_response(self, func_input, func_output):
        self_link = "/%s" % self.model_name
        if self.query_schema and func_input:
            self_link += '?' + self.query_schema.to_json(func_input)

        body = {
            "_links": {
                "self": {"href": self_link}
            },
            "_embedded": {}
        }

        if self.list_metadata:
            l, meta = func_output
            meta = Struct(self.list_metadata).to_json(meta)
            body.update(meta)
        else:
            l = func_output

        body["_embedded"][self.model_name] = []
        for inst in l:
            jrep = Representation(Model(self.full_model_name)).to_json(inst)
            body["_embedded"][self.model_name].append(jrep)

        return Response(json.dumps(body), 200, {"Content-Type": "application/json"})





from unittest2 import TestCase
from multiprocessing import Process

from werkzeug.wrappers import Response
from werkzeug.test import Client as TestClient

from cosmic.http import Server
from cosmic.globals import cosmos
from cosmic.client import APIClient, ClientLoggingMixin
from cosmic.testing import served_api, Wildcard
from cosmic.types import *
from words import words


class TestWords(TestCase):

    def setUp(self):
        self._old_cosmos = cosmos.data
        cosmos.data = {}

        app = Server(words).wsgi_app
        self.client = TestClient(app, response_wrapper=Response)

    def tearDown(self):
        cosmos.data = self._old_cosmos

    def test_pluralize_okay(self):
        resp = self.client.post('/actions/pluralize',
                                content_type='application/json', data='"pony"')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data, '"ponies"')


class TestWordsSystem(TestCase):

    def test_single_thread(self):
        with served_api(words, 5003):

            class WordsClient(APIClient):
                base_url = 'http://127.0.0.1:5003'

            c = WordsClient()
            p = Process(target=c.actions.lock_thread, args=(1,))
            p.start()
            self.assertEqual(c.actions.count_letters((None, {"letters": "rabbit"})), 6)
            p.terminate()

    def test_multi_thread(self):
        with served_api(words, 5002, threaded=True):

            class WordsClient(APIClient):
                base_url = 'http://127.0.0.1:5002'

            c = WordsClient()
            p = Process(target=c.actions.lock_thread, args=(1,))
            p.start()
            self.assertEqual(c.actions.count_letters((None, {"letters": "rabbit"})), 6)
            p.terminate()

    def test_system(self):
        with served_api(words, 5000):

            class WordsClient(APIClient):
                base_url = 'http://127.0.0.1:5000'

            c = WordsClient()
            self.assertEqual(APISpec.to_json(c.spec), APISpec.to_json(words.spec))

    def test_logging(self):
        with served_api(words, 5001):

            class WordsClient(ClientLoggingMixin, APIClient):
                base_url = 'http://127.0.0.1:5001'

            c = WordsClient()
            c.actions.pluralize('pencil')
            self.assertEqual(c.log[-1][0], {
                'method': 'POST',
                'data': '"pencil"',
                'headers': [('Content-Type', 'application/json')],
                'url': u'/actions/pluralize'
            })
            self.assertEqual(c.log[-1][1], {
                'status_code': 200,
                'data': u'"pencils"',
                'headers': [
                    ('Content-Length', '9'),
                    ('Content-Type', 'application/json'),
                    ('Date', Wildcard),
                    ('Server', Wildcard),
                ]
            })








__all__ = ['BaseModel']


class BaseModel(object):
    """Subclasses of this class are fed into the :meth:`~cosmic.api.API.model`
    decorator to attach models to an API. The API object doesn't care about
    how you implement this class, it just copies the necessary properties and
    leaves the class alone.
    """
    #: A list of properties which, along with
    #: :data:`~cosmic.models.BaseModel.links` below, will be used for the
    #: model's representation and patch, defined in the same way Teleport
    #: Struct fields are defined:
    #:
    #: .. code:: python
    #:
    #:     properties = [
    #:         required('name', String),
    #:         optional('age', Integer),
    #:     ]
    #:
    #: See :class:`~cosmic.types.Representation` and
    #: :class:`~cosmic.types.Patch`.
    properties = []
    #: A list of methods that this model supports. Possible values are
    #: ``'get_by_id'``, ``'create'``, ``'update'``, ``'delete'`` and
    #: ``'get_list'``.
    methods = []
    #: A list of properties for the :meth:`get_list` handler. They are defined
    #: in the same way as :data:`properties` above.
    query_fields = []
    #: A list of properties that can be returned along with the usual response
    #: for :meth:`~cosmic.models.BaseModel.get_list`. These can be used for
    #: things like pagination.
    list_metadata = []
    #: Similar to properties, but encodes a relationship between this model
    #: and another. In database terms this would be a foreign key. Use
    #: :func:`~cosmic.types.required_link` and
    #: :func:`~cosmic.types.optional_link` to specify them.
    links = []

    @classmethod
    def get_by_id(cls, id):
        """
        :param id:
        :return: Model representation
        :raises cosmic.exceptions.NotFound:
        """
        raise NotImplementedError()

    @classmethod
    def get_list(cls, **kwargs):
        """
        :param kwargs: Defined by \
            :data:`~cosmic.models.BaseModel.query_fields`
        :return: If model does not define \
            :data:`~cosmic.models.BaseModel.list_metadata`, returns a list of
            tuples of models ids and representations. Otherwise returns a
            tuple where the first element is the above list, and the second is
            a dict as specified by \
            :data:`~cosmic.models.BaseModel.list_metadata`.
        """
        raise NotImplementedError()

    @classmethod
    def create(cls, **valid_patch):
        """
        :param validated_patch: The model patch.
        :return: A tuple of model id and model representation.
        """
        raise NotImplementedError()

    @classmethod
    def update(cls, id, **valid_patch):
        """
        :param id:
        :param validated_patch:
        :return: The model representation after patch has been applied.
        :raises cosmic.exceptions.NotFound:
        """
        raise NotImplementedError()

    @classmethod
    def delete(cls, id):
        """
        :param id:
        :raises cosmic.exceptions.NotFound:
        """
        raise NotImplementedError()

    @classmethod
    def validate_patch(cls, patch):
        """
        :param patch: The model patch
        :raises cosmic.exceptions.ValidationError:

        Run before any :meth:`~cosmic.models.BaseModel.create` or
        :meth:`~cosmic.models.BaseModel.update` call to validate the patch.
        All fields are made optional for the patch, so this method is a chance
        to ensure that the expected values were indeed passed in.
        """
        pass




import datetime

from cosmic.api import API
from cosmic.exceptions import HTTPError
from cosmic.testing import DBModel, db
from cosmic.types import *
from cosmic.http import Server


planet_db = {
    "Sphere": {
        "0": {
            u"name": u"Sun",
            u"temperature": 3000.0,
        },
        "1": {
            u"name": u"Earth",
            u"temperature": 30.0,
            u"revolves_around": "0",
        },
        "2": {
            u"name": u"Moon",
            u"temperature": -50.0,
            u"revolves_around": "1",
        },
    }
}

planetarium = API("planetarium")


class PlanetariumServer(Server):
    def parse_request(self, endpoint, request, **url_args):
        if request.headers.get('X-Danish', None) != "poppyseed":
            raise HTTPError(401, "Unauthorized")
        return super(PlanetariumServer, self).parse_request(endpoint, request, **url_args)


planetarium_server = PlanetariumServer(planetarium)


@planetarium.model
class Sphere(DBModel):
    table_name = 'Sphere'
    methods = ["get_list", "get_by_id", "create", "update", "delete"]

    properties = [
        required("name", String),
        required("temperature", Float),
    ]
    links = [
        optional_link("revolves_around", Model('planetarium.Sphere')),
    ]
    query_fields = [
        optional("name", String),
        optional("revolves_around", String)
    ]
    list_metadata = [
        required('last_updated', DateTime)
    ]

    @classmethod
    def get_list(cls, **kwargs):
        l = super(Sphere, cls).get_list(**kwargs)
        metadata = {
            'last_updated': datetime.datetime(year=2014, month=1, day=1)
        }
        return (l, metadata)

    @classmethod
    def validate_patch(cls, datum):
        if 'temperature' in datum:
            raise ValidationError("Temperature is readonly")
        if "name" not in datum:
            raise ValidationError("Name is required")
        if datum["name"][0].islower():
            raise ValidationError("Name must be capitalized", datum["name"])

    @classmethod
    def create(cls, **patch):
        return super(Sphere, cls).create(temperature=60, **patch)


@planetarium.action(accepts=Representation(Model('planetarium.Sphere')), returns=String)
def hello(sphere):
    return "Hello, %s" % sphere[1]['name']


if __name__ == "__main__":
    with db.swap(planet_db):
        planetarium.run()



from __future__ import unicode_literals


class Either(object):
    def __init__(self, exception=None, value=None):
        if exception is not None and value is not None:
            raise RuntimeError("Impossible Either")
        self.exception = exception
        self.value = value


class SpecError(Exception):
    pass


class NotFound(Exception):
    """Expected to be raised by :meth:`get_by_id`, :meth:`update` and
    :meth:`delete` when the resource is not found. Cosmic will convert
    it to a 404 response on the server, and on the client, it will
    interpret this response by reraising the exception.
    """
    pass


class HTTPError(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message
        super(HTTPError, self).__init__(code, message)


class RemoteHTTPError(HTTPError):
    pass


class ThreadLocalMissing(Exception):
    """Raised when trying to access a value inside
    :class:`cosmic.globals.ThreadLocalDict` without creating a thread-local
    to store it on. See :func:`cosmic.globals.thread_local` and
    :func:`cosmic.globals.thread_local_middleware`.
    """
    pass



import json
import copy

import requests
from requests.sessions import Session
from requests.structures import CaseInsensitiveDict

from werkzeug.test import Client as WerkzeugTestClient
from werkzeug.wrappers import Response

from .api import BaseAPI, Object
from .types import *
from .globals import cosmos
from .http import CreateEndpoint, DeleteEndpoint, GetByIdEndpoint, \
    GetListEndpoint, UpdateEndpoint, ActionEndpoint, SpecEndpoint


class BaseAPIClient(BaseAPI):

    def __init__(self, *args, **kwargs):
        super(BaseAPIClient, self).__init__(*args, **kwargs)
        self._generate_handler_objects()

    def call(self, endpoint, *args, **kwargs):
        req = self.build_request(endpoint, *args, **kwargs)
        res = self.make_request(endpoint, req)
        return self.parse_response(endpoint, res)

    def build_request(self, endpoint, *args, **kwargs):
        return endpoint.build_request(*args, **kwargs)

    def parse_response(self, endpoint, res):
        return endpoint.parse_response(res)

    def make_request(self, endpoint, request):
        raise NotImplementedError()

    def _generate_handler_objects(self):
        from functools import partial

        spec = self.spec

        for name, action in spec["actions"].items():
            setattr(self.actions, name,
                    partial(self.call, ActionEndpoint(spec, name)))

        for name, modeldef in spec["models"].items():
            m = Object()
            m.create = partial(self.call, CreateEndpoint(spec, name))
            m.update = partial(self.call, UpdateEndpoint(spec, name))
            m.delete = partial(self.call, DeleteEndpoint(spec, name))
            m.get_list = partial(self.call, GetListEndpoint(spec, name))
            m.get_by_id = partial(self.call, GetByIdEndpoint(spec, name))
            m.validate_patch = lambda patch: None

            setattr(self.models, name, m)


class APIClient(BaseAPIClient):
    verify = True
    base_url = None

    def __init__(self, *args, **kwargs):
        self.session = Session()
        spec = self.call(SpecEndpoint())
        super(APIClient, self).__init__(*args, spec=spec, **kwargs)

    def make_request(self, endpoint, request):
        request.url = self.base_url + request.url
        prepared = self.session.prepare_request(request)
        return self.session.send(prepared,
                                 stream=False,
                                 timeout=None,
                                 verify=self.verify,
                                 cert=None,
                                 proxies={},
                                 allow_redirects=True)


class WsgiAPIClient(BaseAPIClient):
    wsgi_app = None
    server_cosmos = None

    def __init__(self, *args, **kwargs):
        self.client = WerkzeugTestClient(self.wsgi_app, response_wrapper=Response)
        spec = self.call(SpecEndpoint())
        super(WsgiAPIClient, self).__init__(*args, spec=spec, **kwargs)

    def make_request(self, endpoint, request):
        kwargs = {
            "method": request.method,
            "data": request.data,
            "headers": request.headers
        }
        # Content-Type should be provided as kwarg because otherwise we can't
        # access request.mimetype
        if 'Content-Type' in request.headers:
            kwargs['content_type'] = request.headers.pop('Content-Type')

        if self.server_cosmos is None:
            r = self.client.open(path=request.url, **kwargs)
        else:
            with cosmos.swap(self.server_cosmos):
                r = self.client.open(path=request.url, **kwargs)


        resp = requests.Response()
        resp._content = r.data
        resp.headers = CaseInsensitiveDict(r.headers)
        resp.status_code = r.status_code

        return resp


class ClientLoggingMixin(object):

    def __init__(self, *args, **kwargs):
        self.log = []
        self.__last_request = None
        super(ClientLoggingMixin, self).__init__(*args, **kwargs)

    def build_request(self, endpoint, *args, **kwargs):
        request = super(ClientLoggingMixin, self).build_request(endpoint, *args, **kwargs)
        self.__last_request = {
            "method": request.method,
            "data": request.data,
            "headers": copy.deepcopy(request.headers.items()),
            "url": request.url,
        }
        return request

    def parse_response(self, endpoint, res):
        headers = []
        for key in sorted(res.headers.keys()):
            headers.append((key.title(), res.headers[key]))
        saved_resp = {
            "data": res.text,
            "headers": headers,
            "status_code": res.status_code
        }
        self.log.append((self.__last_request, saved_resp))
        return super(ClientLoggingMixin, self).parse_response(endpoint, res)






#!/usr/bin/python"""
"""
Cosmic
------

Cosmic is a tiny web API framework based on
`Teleport <http://teleport-json.org/>`_.

Links
`````

* `Website <http://www.cosmic-api.com/>`_
* `Documentation <http://www.cosmic-api.com/docs/cosmic/python/latest/>`_
* `Development version <https://github.com/cosmic-api/cosmic.py>`_

"""

from setuptools import setup

import os
root = os.path.abspath(os.path.dirname(__file__))

with open("{}/LICENSE".format(root)) as license_file:
    license_text = license_file.read()
with open("{}/CHANGES.rst".format(root)) as changes_file:
    changelog_text = changes_file.read()

long_description = """
{}

License
```````

{}

Changelog
`````````

{}
""".format(__doc__, license_text, changelog_text)

setup(
    name='cosmic',
    version="0.5.5",
    url='http://www.cosmic-api.com/docs/cosmic/python/',
    packages=['cosmic'],
    description='A tiny web API framework based on Teleport',
    license='MIT',
    author='8313547 Canada Inc.',
    author_email='alexei@boronine.com',
    long_description=long_description,
    install_requires=[
        'Werkzeug>=0.9.1',
        'requests>=2.2.0',
        'isodate>=0.5.1',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
)



from __future__ import unicode_literals
import json
import inspect
from collections import OrderedDict

from .tools import get_args, assert_is_compatible, \
    validate_underscore_identifier
from .types import *
from .globals import cosmos
from . import MODEL_METHODS


class Object(object):
    pass


class BaseAPI(object):
    """This class represents the common interface for the server-side and the
    client-side API objects. It contains the API spec as well as a collection
    of functions for each API endpoint. On the server side, these functions
    are user-defined. On the client side, they are created automatically to
    execute HTTP calls.
    """

    def __init__(self, spec):
        #: The API spec, in its native form (see
        #: :class:`~cosmic.types.APISpec`)
        self.spec = spec
        #: This stores the action functions as properties of a plain object:
        #:
        #: .. code:: python
        #:
        #:     >>> spelling.actions.correct('simpl')
        #:     "simple"
        #:
        self.actions = Object()
        #: Through this property you can access the model functions
        #: :meth:`get_by_id`, :meth:`get_list`, :meth:`create`, :meth:`update`
        #: and :meth:`delete`. They are accessed as attributes:
        #:
        #: .. code:: python
        #:
        #:     >>> quotes.models.Quote.get_by_id("1")
        #:     {"text": "Know thyself.", "author": "Socrates"}
        #:
        self.models = Object()
        name = self.spec['name']
        if name in cosmos.keys():
            raise RuntimeError("API already exists: {}".format(name))
        cosmos[name] = self



class API(BaseAPI):
    """
    :param name: The API name is required, and should be unique
    :param homepage: If you like, the API spec may include a link to your
        homepage
    """

    def __init__(self, name, homepage=None):
        super(API, self).__init__({
            "name": name,
            "homepage": homepage,
            "actions": OrderedDict(),
            "models": OrderedDict(),
        })

    def run(self, port=5000, debug=False, **kwargs):
        """Simple way to run the API in development. The debug parameter gets
        passed into a :class:`~cosmic.http.Server` instance, all other
        parameters - into Werkzeug's :func:`~werkzeug.serving.run_simple`. For
        more serving options, see :ref:`guide-serving`.
        """
        from werkzeug.serving import run_simple
        from .http import Server

        server = Server(self, debug=debug)
        run_simple('127.0.0.1', port, server.wsgi_app, **kwargs)


    def action(self, accepts=None, returns=None):
        """A decorator for registering actions with API.

        The *accepts* parameter is a schema that describes the input of the
        function, *returns* is a schema that describes the output of the
        function. The name of the function becomes the name of the action and
        the docstring serves as the action's documentation.

        Once registered, an action will become accessible as an attribute of
        the :data:`~cosmic.api.BaseAPI.actions` object.

        .. code:: python

            >>> random = API("random")
            >>> @random.action(returns=Integer)
            ... def generate():
            ...     "Random enough"
            ...     return 9
            >>> random.actions.generate()
            9

        """

        def wrapper(func):
            name = unicode(func.__name__)
            validate_underscore_identifier(name)
            required_args, optional_args = get_args(func)

            if accepts:
                assert_is_compatible(accepts, required_args, optional_args)

            doc = inspect.getdoc(func)
            self.spec['actions'][name] = {
                "accepts": accepts,
                "returns": returns,
                "doc": doc,
            }

            setattr(self.actions, name, func)

            return func

        return wrapper

    def model(self, model_cls):
        """A decorator for registering a model with an API. The name of the
        model class is used as the name of the resulting model. A subclass
        of :class:`~cosmic.models.BaseModel` is used to supply the necessary
        metadata and functions to the API.

        .. code:: python

            >>> dictionary = API("dictionary")
            >>> @dictionary.model
            ... class Word(BaseModel):
            ...    properties = [
            ...        required("text", String)
            ...    ]
            ...

        Once registered, a model will become accessible as an attribute of the
        :data:`~cosmic.api.BaseAPI.models` object.
        """

        name = model_cls.__name__

        m = Object()
        m.validate_patch = model_cls.validate_patch

        methods = {}
        for method in MODEL_METHODS:
            methods[method] = method in model_cls.methods
            setattr(m, method, getattr(model_cls, method))

        self.spec['models'][unicode(name)] = {
            "properties": OrderedDict(model_cls.properties),
            "links": OrderedDict(model_cls.links),
            "query_fields": OrderedDict(model_cls.query_fields),
            "list_metadata": OrderedDict(model_cls.list_metadata),
            "methods": methods,
        }
        APISpec.assemble(self.spec)
        setattr(self.models, name, m)

        return model_cls

